@@ -13,6 +13,7 @@
 #include "kglobalshortcutinfo_p.h"
 #include "kserviceactioncomponent.h"
 #include "logging_p.h"
+#include "sequencehelpers_p.h"
 #include <config-kglobalaccel.h>
 
 #include <KApplicationTrader>
@@ -158,10 +159,8 @@ void GlobalShortcutsRegistry::migrateKHotkeys()
 }
 
 /*
- * Migrate the config for service actions to a new format that only stores the actual shortcut if not default.
+ * Migrate the Plasma 5 config for service actions to a new format that only stores the actual shortcut if not default.
  * All other information is read from the desktop file.
- * Keep the old data for compatibility with KF5-based kglobalaccel.
- * Once Plasma 6 settles down consider dropping this data
  */
 void GlobalShortcutsRegistry::migrateConfig()
 {
@@ -203,7 +202,7 @@ void GlobalShortcutsRegistry::migrateConfig()
         KDesktopFile file(fileName);
         const QString componentName = QFileInfo(fileName).fileName();
 
-        auto migrateTo = [this, componentName](KConfigGroup &group, const QString &actionName, const QString displayName) {
+        auto migrateTo = [this, componentName](KConfigGroup &group, const QString &actionName) {
             QString migrateFrom = group.readEntry<QString>(QStringLiteral("X-KDE-Migrate-Shortcut"), QString());
 
             if (migrateFrom.isEmpty()) {
@@ -236,12 +235,12 @@ void GlobalShortcutsRegistry::migrateConfig()
         };
 
         KConfigGroup desktopGroup = file.desktopGroup();
-        migrateTo(desktopGroup, QStringLiteral("_launch"), file.readName());
+        migrateTo(desktopGroup, QStringLiteral("_launch"));
 
         const QStringList actions = file.readActions();
         for (const QString &action : actions) {
             KConfigGroup actionGroup = file.actionGroup(action);
-            migrateTo(actionGroup, action, file.actionGroup(action).readEntry<QString>("Name", QString()));
+            migrateTo(actionGroup, action);
         }
     }
 
@@ -392,35 +391,57 @@ GlobalShortcutsRegistry *GlobalShortcutsRegistry::self()
 
 static void correctKeyEvent(int &keyQt)
 {
-    // When we are provided just a Shift key press, interpret it as "Shift" not as "Shift+Shift"
-    switch (keyQt) {
-    case (Qt::ShiftModifier | Qt::Key_Shift).toCombined():
-        keyQt = Qt::Key_Shift;
+    int keyMod = keyQt & Qt::KeyboardModifierMask;
+    int keySym = keyQt & ~Qt::KeyboardModifierMask;
+    switch (keySym) {
+    case Qt::Key_Super_L:
+    case Qt::Key_Super_R:
+        keySym = Qt::Key_Meta;
+        keyMod |= Qt::MetaModifier;
         break;
-    case (Qt::ControlModifier | Qt::Key_Control).toCombined():
-        keyQt = Qt::Key_Control;
-        break;
-    case (Qt::AltModifier | Qt::Key_Alt).toCombined():
-        keyQt = Qt::Key_Alt;
-        break;
-    case (Qt::MetaModifier | Qt::Key_Meta).toCombined():
-        keyQt = Qt::Key_Meta;
+    case Qt::Key_SysReq:
+        // Known limitation:
+        //     When shortcut is Mod(s)+Alt+Print, only works when Alt is released before Mod(s),
+        //     Does not work with multikey shortcuts.
+        // When the user presses Mod(s)+Alt+Print, the SysReq event is fired only
+        // when the Alt key is released. Before we get the Mod(s)+SysReq event, we
+        // first get a Mod(s)+Alt event, breaking multikey shortcuts.
+        keySym = Qt::Key_Print;
+        keyMod |= Qt::AltModifier;
         break;
     }
-    // Known limitation:
-    //     When shortcut is Mod(s)+Alt+Print, only works when Alt is released before Mod(s),
-    //     Does not work with multikey shortcuts.
-    // When the user presses Mod(s)+Alt+Print, the SysReq event is fired only
-    // when the Alt key is released. Before we get the Mod(s)+SysReq event, we
-    // first get a Mod(s)+Alt event, breaking multikey shortcuts.
-    if ((keyQt & ~Qt::KeyboardModifierMask) == Qt::Key_SysReq) {
-        keyQt = Qt::Key_Print | (keyQt & Qt::KeyboardModifierMask) | Qt::AltModifier;
-    }
+    keyQt = keySym | keyMod;
 }
 
 bool GlobalShortcutsRegistry::keyPressed(int keyQt)
 {
     correctKeyEvent(keyQt);
+    const int key = keyQt & ~Qt::KeyboardModifierMask;
+    const Qt::KeyboardModifiers modifiers = static_cast<Qt::KeyboardModifiers>(keyQt & Qt::KeyboardModifierMask);
+    switch (key) {
+    case 0:
+        // Invalid key code
+        m_state = Normal;
+        _active_sequence = QKeySequence();
+        return false;
+    case Qt::Key_Shift:
+    case Qt::Key_Control:
+    case Qt::Key_Alt:
+    case Qt::Key_Super_L:
+    case Qt::Key_Super_R:
+    case Qt::Key_Meta:
+        m_state = PressingModifierOnly;
+        m_currentModifiers = Utils::keyToModifier(key) | modifiers;
+        return false;
+    default:
+        m_state = Normal;
+        m_currentModifiers = modifiers;
+        return processKey(keyQt);
+    }
+}
+
+bool GlobalShortcutsRegistry::processKey(int keyQt)
+{
     int keys[maxSequenceLength] = {0, 0, 0, 0};
     int count = _active_sequence.count();
     if (count == maxSequenceLength) {
@@ -456,7 +477,7 @@ bool GlobalShortcutsRegistry::keyPressed(int keyQt)
         }
     }
 
-    qCDebug(KGLOBALACCELD) << "Pressed key" << QKeySequence(keyQt).toString() << ", current sequence" << _active_sequence.toString() << "="
+    qCDebug(KGLOBALACCELD) << "Processed key" << QKeySequence(keyQt).toString() << ", current sequence" << _active_sequence.toString() << "="
                            << (shortcut ? shortcut->uniqueName() : QStringLiteral("(no shortcut found)"));
     if (!shortcut) {
         // This can happen for example with the ALT-Print shortcut of kwin.
@@ -497,11 +518,59 @@ bool GlobalShortcutsRegistry::keyPressed(int keyQt)
 
 bool GlobalShortcutsRegistry::keyReleased(int keyQt)
 {
-    Q_UNUSED(keyQt)
+    correctKeyEvent(keyQt);
+    bool handled = false;
+    const int key = keyQt & ~Qt::KeyboardModifierMask;
+    const Qt::KeyboardModifiers modifiers = static_cast<Qt::KeyboardModifiers>(keyQt & Qt::KeyboardModifierMask);
+    switch (key) {
+    case 0:
+        // Invalid key code
+        m_state = Normal;
+        break;
+    case Qt::Key_Super_L:
+    case Qt::Key_Super_R:
+    case Qt::Key_Meta:
+    case Qt::Key_Shift:
+    case Qt::Key_Control:
+    case Qt::Key_Alt: {
+        constexpr auto releaseTimeout = std::chrono::milliseconds(200);
+        const auto currentTime = std::chrono::steady_clock::now().time_since_epoch();
+        if (m_state == PressingModifierOnly) {
+            m_modifierOnlyModifiers = m_currentModifiers;
+            m_state = ReleasingModifierOnly;
+            m_modifierFirstReleaseTime = currentTime;
+        }
+        m_currentModifiers = modifiers & ~Utils::keyToModifier(key);
+        if (m_state == ReleasingModifierOnly && !m_currentModifiers) {
+            m_state = Normal;
+            if (currentTime - m_modifierFirstReleaseTime < releaseTimeout) {
+                handled = processKey(m_modifierOnlyModifiers);
+            }
+        }
+        break;
+    }
+    default:
+        m_state = Normal;
+        break;
+    }
     if (m_lastShortcut) {
         m_lastShortcut->context()->component()->emitGlobalShortcutReleased(*m_lastShortcut);
         m_lastShortcut = nullptr;
     }
+    return handled;
+}
+
+bool GlobalShortcutsRegistry::pointerPressed(Qt::MouseButtons pointerButtons)
+{
+    Q_UNUSED(pointerButtons)
+    m_state = Normal;
+    return false;
+}
+
+bool GlobalShortcutsRegistry::axisTriggered(int axis)
+{
+    Q_UNUSED(axis)
+    m_state = Normal;
     return false;
 }
 
